package com.smppw.analysis.application.service.position.stock;

import cn.hutool.core.collection.CollUtil;
import cn.hutool.core.collection.ListUtil;
import com.smppw.analysis.application.dto.position.*;
import com.smppw.analysis.application.dto.position.stock.RiskExposureParams;
import com.smppw.analysis.application.dto.position.stock.RiskExposureVO;
import com.smppw.analysis.application.service.position.AbstractNonSynthesizeBizHandler;
import com.smppw.analysis.application.service.position.BizHandlerConstants;
import com.smppw.analysis.application.service.position.PositionLoadFactory;
import com.smppw.analysis.domain.gateway.CacheFactory;
import com.smppw.analysis.infrastructure.config.AnalysisProperty;
import com.smppw.analysis.infrastructure.exception.APIException;
import com.smppw.constants.SecType;
import lombok.AllArgsConstructor;
import lombok.Getter;
import org.springframework.stereotype.Component;

import java.math.BigDecimal;
import java.math.RoundingMode;
import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.function.Predicate;
import java.util.stream.Collectors;

/**
 * @author wangzaijun
 * @date 2023/6/13 8:49
 * @description 私募基金股票风险敞口走势
 */
@Component(BizHandlerConstants.RISK_EXPOSURE)
public class RiskExposureBizHandler extends AbstractNonSynthesizeBizHandler<RiskExposureParams, List<RiskExposureVO>> {
    private static final List<RiskExposure> RISK_EXPOSURES = ListUtil.list(true);

    static {
        // 个股多头
        RISK_EXPOSURES.add(new RiskExposure(RiskExposureEnum.STOCK_BULL, true,
                e -> AssetCategoryEnum.STOCK.name().equals(e.getAsset().getValue())
                        && e.getSharesNature() == 1 && e.getLevel() >= 4));
        // 个股空头
        RISK_EXPOSURES.add(new RiskExposure(RiskExposureEnum.STOCK_BEAR, false,
                e -> AssetCategoryEnum.STOCK.name().equals(e.getAsset().getValue())
                        && e.getSharesNature() == 2 && e.getLevel() >= 4));
        // 股指多头
        RISK_EXPOSURES.add(new RiskExposure(RiskExposureEnum.FUTURE_BULL, true,
                e -> AssetCategoryEnum.FUTURE.name().equals(e.getAsset().getValue())
                        && e.getSharesNature() == 1
                        && e.getSubType() == 31
                        && e.getLevel() >= 4));
        // 股指空头
        RISK_EXPOSURES.add(new RiskExposure(RiskExposureEnum.FUTURE_BEAR, false,
                e -> AssetCategoryEnum.FUTURE.name().equals(e.getAsset().getValue())
                        && e.getSharesNature() == 2
                        && e.getSubType() == 31
                        && e.getLevel() >= 4));
        // ETF融券
        RISK_EXPOSURES.add(new RiskExposure(RiskExposureEnum.ETF, false,
                e -> e.getSubjectCode() != null && e.getSubjectCode().contains("2101")
                        && e.getLevel() == 3
                        && e.getRef() != null && e.getRef().getLabel() != null && e.getRef().getLabel().contains("ETF")));
    }

    public RiskExposureBizHandler(AnalysisProperty property, CacheFactory cacheFactory, PositionLoadFactory factory) {
        super(property, cacheFactory, factory);
    }

    @Override
    protected List<RiskExposureVO> afterNonSynthesizeLoadData(RiskExposureParams params, List<FundPositionBaseInfo> baseInfos, List<FundPositionDetail> positionInfos) {
        if (CollUtil.isEmpty(positionInfos)) {
            return ListUtil.empty();
        }
        List<String> types = ListUtil.of(SecType.PRIVATE_FUND, SecType.PRIVATELY_OFFERED_FUND);
        String fundType = positionInfos.get(0).getFundType();
        if (!types.contains(fundType)) {
            throw new APIException("股票风险敞口接口仅支持私募基金和自建基金！");
        }
        List<RiskExposureVO> resultList = ListUtil.list(false);
        // 按日期分组处理
        Map<String, List<FundPositionDetail>> collect = positionInfos.stream().collect(Collectors.groupingBy(FundPositionDetail::getDate));
        collect.forEach((k, v) -> {
            BigDecimal total = baseInfos.stream().filter(e -> k.equals(e.getDate())).map(FundPositionBaseInfo::getAssetNv)
                    .filter(Objects::nonNull).findFirst().map(BigDecimal::abs).orElse(null);
            if (total != null && total.abs().compareTo(BigDecimal.ZERO) > 0) {
                List<RefMarketValueRatio> mvrs = ListUtil.list(true);
                BigDecimal exposureMv = BigDecimal.ZERO;
                // 是否全部都null，
                boolean allNullFlag = true;
                for (RiskExposure riskExposure : RISK_EXPOSURES) {
                    BigDecimal mv = v.stream().filter(riskExposure.getPredicate()).map(FundPositionDetail::getMarketValue)
                            .filter(Objects::nonNull).reduce(BigDecimal::add).orElse(null);
                    BigDecimal ratio = null;
                    if (mv != null) {
                        ratio = mv.divide(total, 4, RoundingMode.HALF_UP);
                        exposureMv = riskExposure.operatePlus ? exposureMv.add(mv.abs()) : exposureMv.subtract(mv.abs());
                        allNullFlag = false;
                    }
                    mvrs.add(new RefMarketValueRatio(RiskExposureEnum.getValueLabel(riskExposure.exposureEnum), mv, ratio));
                }
                // 敞口 市值不为null才添加到结果集中
                if (!allNullFlag) {
                    BigDecimal exposureRatio = exposureMv.divide(total, 4, RoundingMode.HALF_UP);
                    mvrs.add(new RefMarketValueRatio(RiskExposureEnum.getValueLabel(RiskExposureEnum.EXPOSURE), null, exposureRatio));
                    RiskExposureVO vo = new RiskExposureVO();
                    vo.setDate(k);
                    vo.setMvr(mvrs);
                    resultList.add(vo);
                }
            }
        });
        if (CollUtil.isNotEmpty(resultList)) {
            resultList.sort(Comparator.comparing(RiskExposureVO::getDate).reversed());
        }
        return resultList;
    }

    @Getter
    @AllArgsConstructor
    private static class RiskExposure {
        /**
         * 类型
         */
        private final RiskExposureEnum exposureEnum;
        /**
         * 是否加法
         */
        private final boolean operatePlus;
        /**
         * 数据过滤条件
         */
        private final Predicate<FundPositionDetail> predicate;
    }
}
